/******************************************************************************
 * $Id: hugefileresolver.cpp 22205 2011-04-18 21:17:30Z rouault $
 *
 * Project:  GML Reader
 * Purpose:  Implementation of GMLReader::HugeFileResolver() method.
 * Author:   Alessandro Furieri, a.furitier@lqt.in
 *
 ******************************************************************************
 * Copyright (c) 2011, Alessandro Furieri
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 *
 ******************************************************************************
 * Contributor: Alessandro Furieri, a.furieri@lqt.it
 * This module implents GML_SKIP_RESOLVE_ELEMS HUGE
 * Developed for Faunalia ( http://www.faunalia.it) with funding from 
 * Regione Toscana - Settore SISTEMA INFORMATIVO TERRITORIALE ED AMBIENTALE
 *
 ****************************************************************************/

#include "gmlreader.h"
#include "cpl_error.h"

CPL_CVSID("$Id: hugefileresolver.cpp 22205 2011-04-18 21:17:30Z rouault $");

#include "gmlreaderp.h"
#include "gmlutils.h"
#include "cpl_conv.h"
#include "cpl_string.h"
#include "cpl_http.h"

#include <stack>

/*
/ an internal helper struct supporting GML tags
/ <Node>, <Edge>, <Face> and similar 
*/
struct huge_file_tag_helper
{
    char szTagName[1024];
    char szIdHref[1024];
    char szNodeFrom[1024];
    char szNodeTo[1024];
    int bIsGmlHref;
    int iOffsetStart;
    int iOffsetEnd;
    int bHasCoords;
    int bHasZ;
    int bIsNodeGmlId;
    double xNodeFrom;
    double yNodeFrom;
    double zNodeFrom;
    double xNodeTo;
    double yNodeTo;
    double zNodeTo;
    struct huge_file_tag_helper *pPrev;
    struct huge_file_tag_helper *pNext;
};

/*
/ an internal helper struct supporting GML 
/ resolver for Huge Files (based on SQLite)
*/
struct huge_file_helper
{
    sqlite3 *hDB;
    sqlite3_stmt *hNodes;
    sqlite3_stmt *hStrings;
    sqlite3_stmt *hHRefs;
    int iItemId;
    int bFromNode;
    char szNodeFrom[1024];
    char szNodeTo[1024];
    char szNodeSrs[1024];
    struct huge_file_tag_helper *pFirst;
    struct huge_file_tag_helper *pLast;
};

/* an internal helper struct for GML rewriting */
struct huge_file_sub_resolver
{
    char *pszString;
    char *pszBefore;
    char *pszTagOpen;
    char *pszTagClose;
    int iOffsetStart;
    int iOffsetEnd;
    struct huge_file_sub_resolver *pNext;
};

/* an internal helper struct for GML rewriting */
struct huge_file_resolver
{
    int iRowID;
    int iItemID;
    char *pszUnresolved;
    char *pszAfter;
    char *pszResolved;
    struct huge_file_sub_resolver *pFirst;
    struct huge_file_sub_resolver *pLast;
    struct huge_file_sub_resolver **pArray;
};

/* an internal helper struct supporting Feature Class */
struct huge_class_item
{
    char        *m_pszName;
    int          n_nNameLen;
    int          n_nItemCount;
    int          n_nGeomCount;
    struct       huge_class_item *pNext;
};

/* an internal helper struct supporting Feature Classes */
struct huge_class_container
{
    int          m_bSequentialLayers;
    struct       huge_class_item *pFirst;
    struct       huge_class_item *pLast;
};

static struct huge_class_container * gmlAllocClassContainer( void )
{
/* allocating and initializing the ClassContainer */
    struct huge_class_container *pCC = (struct huge_class_container *)
                              CPLMalloc( sizeof (struct huge_class_container) );
    pCC->m_bSequentialLayers = TRUE;
    pCC->pFirst = NULL;
    pCC->pLast = NULL;
    return pCC;
}

static void gmlDestroyClassContainer( struct huge_class_container *pCC )
{
/* destroying the ClassContainer */
    struct huge_class_item *pCI = pCC->pFirst;
    while ( pCI != NULL )
    { 
        CPLFree( pCI->m_pszName );
        CPLFree( pCI );
        pCI = pCI->pNext;
    }
    CPLFree( pCC );
}

static void gmlAddClassToContainer( struct huge_class_container *pCC,
                                    const char *m_pszName )
{
/* adding a further Feature Class */
    struct huge_class_item *pCI = (struct huge_class_item *)
	                           CPLMalloc( sizeof (struct huge_class_item) );
    pCI->m_pszName = CPLStrdup( m_pszName );
    pCI->n_nNameLen = strlen(m_pszName);
    pCI->n_nItemCount = 0;
    pCI->n_nGeomCount = 0;
    pCI->pNext = NULL;	
/* inserting into the linked list */
    if ( pCC->pFirst == NULL )
        pCC->pFirst = pCI;
    if ( pCC->pLast != NULL )
        pCC->pLast->pNext = pCI;
    pCC->pLast = pCI;
}

static void gmlUpdateClass( struct huge_class_container *pCC,
                            const char *m_pszName, int b_has_geom )
{
/* updating the ClassContainer */
    if ( pCC->pFirst == NULL)
    {
    /* empty ClassContainer: first item */
        gmlAddClassToContainer ( pCC, m_pszName );
        pCC->pLast->n_nItemCount ++;
        if ( b_has_geom == TRUE )
            pCC->pLast->n_nGeomCount ++;
        return;
    }
    if( EQUAL(m_pszName, pCC->pLast->m_pszName) )
    {
    /* continuing the same Class group */
        pCC->pLast->n_nItemCount ++;
        if ( b_has_geom == TRUE )
            pCC->pLast->n_nGeomCount ++;
        return;
    }

    struct huge_class_item *pCI = pCC->pFirst;
    while ( pCI != NULL )
    {
        if( EQUAL(m_pszName, pCI->m_pszName) )
        {
        /* Class already declared: NOT SEQUENTIAL */
            pCC->m_bSequentialLayers = FALSE;
            pCI->n_nItemCount ++;
            if ( b_has_geom == TRUE )
                pCC->pLast->n_nGeomCount ++;
            return;
        } 
        pCI = pCI->pNext;
    }

    /* inserting a new Class */
    gmlAddClassToContainer( pCC, m_pszName );
    pCC->pLast->n_nItemCount ++;
    if ( b_has_geom == TRUE )
        pCC->pLast->n_nGeomCount ++;
}

static int gmlHugeFileSQLiteInit( struct huge_file_helper *helper )
{
/* attempting to create SQLite tables */
    const char * osCommand;
    char *pszErrMsg = NULL;
    int rc;
    sqlite3 *hDB = helper->hDB;
    sqlite3_stmt *hStmt;

    /* DB table: NODES */
    osCommand = "CREATE TABLE nodes ("
                "     gml_id VARCHAR PRIMARY KEY, "
                "     x DOUBLE, "
                "     y DOUBLE, "
                "     z DOUBLE, "
                "     srs VARCHAR)";
    rc = sqlite3_exec( hDB, osCommand, NULL, NULL, &pszErrMsg );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to create table nodes: %s",
                  pszErrMsg );
        sqlite3_free( pszErrMsg );
        return FALSE;
    }

    /* DB table: GML_STRINGS */
    osCommand = "CREATE TABLE gml_strings ("
                "     item_id INTEGER, "
                "     gml_id VARCHAR, "
                "     gml_string BLOB, "
                "     gml_resolved BLOB, "
                "     clean_start INTEGER, "
                "     clean_end INTEGER) ";
    rc = sqlite3_exec( hDB, osCommand, NULL, NULL, &pszErrMsg );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to create table gml_strings: %s",
                  pszErrMsg );
        sqlite3_free( pszErrMsg );
        return FALSE;
    }

    /* DB table: GML_HREFS */
    osCommand = "CREATE TABLE gml_hrefs ("
                "     item_id INTEGER, "
                "     gml_href VARCHAR, "
                "     offset_start INTEGER, "
                "     offset_end INTEGER, "
                "     explored INTEGER)";
    rc = sqlite3_exec( hDB, osCommand, NULL, NULL, &pszErrMsg );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to create table gml_hrefs: %s",
                  pszErrMsg );
        sqlite3_free( pszErrMsg );
        return FALSE;
    }

    /* DB table: NODES / Insert cursor */
    osCommand = "INSERT OR IGNORE INTO nodes (gml_id, x, y, z, srs) "
                "VALUES (?, ?, ?, ?, ?)";
    rc = sqlite3_prepare( hDB, osCommand, -1, &hStmt, NULL );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to create INSERT stmt for: nodes" );
        return FALSE;
    }
    helper->hNodes = hStmt;

    /* DB table: GML_STRINGS / Insert cursor */
    osCommand = "INSERT INTO gml_strings "
                "(item_id, gml_id, gml_string, "
                "gml_resolved, clean_start, clean_end) "
                "VALUES (?, ?, ?, ?, ?, ?)";
    rc = sqlite3_prepare( hDB, osCommand, -1, &hStmt, NULL );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to create INSERT stmt for: gml_strings" );
        return FALSE;
    }
    helper->hStrings = hStmt;

    /* DB table: GML_HREFS / Insert cursor */
    osCommand = "INSERT INTO gml_hrefs (item_id, gml_href, offset_start, "
                "offset_end) VALUES (?, ?, ?, ?)";
    rc = sqlite3_prepare( hDB, osCommand, -1, &hStmt, NULL );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to create INSERT stmt for: gml_hrefs" );
        return FALSE;
    }
    helper->hHRefs = hStmt;
		
    /* starting a TRANSACTION */
    rc = sqlite3_exec( hDB, "BEGIN", NULL, NULL, &pszErrMsg );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to perform BEGIN TRANSACTION: %s",
                  pszErrMsg );
        sqlite3_free( pszErrMsg );
        return FALSE;
    }	

    return TRUE;
}

static int gmlHugeFileSQLiteIndex( struct huge_file_helper *helper )
{
/* attempting to create SQLite tables */
    const char * osCommand;
    char *pszErrMsg = NULL;
    int rc;
    sqlite3 *hDB = helper->hDB;

    /* DB index: STRING_GML */
    osCommand = "CREATE INDEX string_gml ON gml_strings (gml_id)";
    rc = sqlite3_exec( hDB, osCommand, NULL, NULL, &pszErrMsg );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to create index string_gml: %s",
                  pszErrMsg );
        sqlite3_free( pszErrMsg );
        return FALSE;
    }

    /* DB index: STRING_ID */
    osCommand = "CREATE INDEX string_id ON gml_strings (item_id)"; 
    rc = sqlite3_exec( hDB, osCommand, NULL, NULL, &pszErrMsg );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to create index string_id: %s",
                  pszErrMsg );
        sqlite3_free( pszErrMsg );
        return FALSE;
    }

    /* DB index: HREF_GML */
    osCommand = "CREATE INDEX href_gml ON gml_hrefs (gml_href)"; 
    rc = sqlite3_exec( hDB, osCommand, NULL, NULL, &pszErrMsg );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to create index href_gml: %s",
                  pszErrMsg );
        sqlite3_free( pszErrMsg );
        return FALSE;
    }

    /* DB index: HREF_ID */
    osCommand = "CREATE INDEX href_id ON gml_hrefs (item_id)";
    rc = sqlite3_exec( hDB, osCommand, NULL, NULL, &pszErrMsg );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to create index href_id: %s",
                  pszErrMsg );
        sqlite3_free( pszErrMsg );
        return FALSE;
    }	

    return TRUE;
}

static void gmlResolverSplitRewrite( const char * pszStr, int iOffsetStart,
                                     int iOffsetEnd, char *szTagOpen, 
                                     char *szTagClose )
{
/* rewriting and splitting a tag */
    int iLen;
    int iEnd;
    int iPos;
    int iPos2;
    int iQuotes = 0;
    char szIn[4196];
    const char *pszTok;
    char *pszOut;
    int bIsGmlDirectedEdge = FALSE;
    int bIsDirectedEdge = FALSE;
    *szTagOpen = '\0';
    *szTagClose = '\0';

    /* extracting the substring to be procecessed */
    iLen = iOffsetEnd - iOffsetStart + 1;
    memcpy ( szIn, pszStr + iOffsetStart, iLen );
    *( szIn + iLen ) = '\0';

    /* checking for gml:directedEdge */
    if ( strstr (szIn, "gml:directedEdge") != NULL)
        bIsGmlDirectedEdge = TRUE;

    if (!bIsGmlDirectedEdge)
    {
    /* checking for directedEdge */
        if ( strstr (szIn, "directedEdge") != NULL)
            bIsDirectedEdge = TRUE;
    }

    /* 
    / extracting the opening tag 
    / excluding the [xlink:href="...."] substring
    */
    pszTok = strstr ( szIn, "xlink:href" );
    iEnd = pszTok - szIn;
    pszOut = szTagOpen;
    for ( int iInd = 0; iInd < iEnd; iInd++)
        *pszOut++ = szIn[iInd];
    for ( int iInd = iEnd; iInd < iLen; iInd++ )
    {
        if ( szIn[iInd] == '"' )
        {
            iQuotes ++;

            /* the token to be suppressed stops here */
            if ( iQuotes == 2 )
            {
                iPos = iInd + 1;
                break;
            }
        }
    }

    /* opening tag: after [xlink:href="...."] */
    iPos2 = iLen;
    for ( int iInd = iPos; iInd < iLen; iInd++ )
    {
        if (bIsGmlDirectedEdge || bIsDirectedEdge)
        {
            if (szIn[iInd] == '/')
                continue;
        }
        if ( szIn[iInd] == '<' )
        {
            iPos2 = iInd;
            break;
        }
        *pszOut++ = szIn[iInd];
    }
    *pszOut++ = '\0';

    /* extracting the closing tag */
    pszOut = szTagClose;
    for ( int iInd = iPos2; iInd < iLen; iInd++)
        *pszOut++ = szIn[iInd];
    *pszOut++ = '\0';
	
    if (strlen(szTagClose) == 0)
    {
        if (bIsGmlDirectedEdge)
        {		
            /* explicit closing tag for gml:directedEdge */
            strcpy(szTagClose, "</gml:directedEdge>");
        }
        if (bIsDirectedEdge)
        {		
            /* explicit closing tag for directedEdge */
            strcpy(szTagClose, "</directedEdge>");
        }
    }
}

static int gmlResolverResolve ( struct huge_file_resolver *resolver )
{
/* attempting to resolve a GML string */
    int iOutLen = 0;
    int iCount = 0;
    int iSub;
    int bSort;
    int iPos;
    int iLen;
    char szTagOpen[2048];
    char szTagClose[2048];
    struct huge_file_sub_resolver *pSub = resolver->pFirst;
    while ( pSub != NULL)
    {
        /* checks for still unresolved dependencies */
        if ( pSub->pszString == NULL )
            return FALSE;
        iOutLen += strlen( pSub->pszString );
        iCount ++;
        pSub = pSub->pNext;
    }	
    if ( resolver->pszUnresolved == NULL || iCount == 0 )
        return FALSE;
	
    /* allocating the sub items array */
    resolver->pArray = (struct huge_file_sub_resolver **)CPLMalloc 
                       ( sizeof (struct huge_file_sub_resolver *) * iCount );

    /* initializing the sub items array */
    iSub = 0;
    pSub = resolver->pFirst;
    while ( pSub != NULL)
    {
        *( resolver->pArray + iSub ) = pSub;
        iSub ++;
        pSub = pSub->pNext;
    }	

    /* sorting the sub items array */
    bSort = TRUE;
    while ( bSort )
    {
        bSort = FALSE;
        for ( iSub = 1; iSub < iCount; iSub++ )
        {
            struct huge_file_sub_resolver *pS1 = *( resolver->pArray + iSub - 1 );
            struct huge_file_sub_resolver *pS2 = *( resolver->pArray + iSub );
            if ( pS1->iOffsetStart > pS2->iOffsetStart )
            {
                *( resolver->pArray + iSub - 1 ) = pS2;
                *( resolver->pArray + iSub ) = pS1;
                bSort = TRUE;
            }
        }
    }	

    /* preparing GML fragments */
    iPos = 0;
    for ( iSub = 0; iSub < iCount; iSub++ )
    {
        pSub = *( resolver->pArray + iSub );
        iLen = pSub->iOffsetStart - iPos;
        if ( iLen > 0 )
        {
            pSub->pszBefore = (char *)CPLMalloc( iLen + 1 );
            memcpy ( pSub->pszBefore, resolver->pszUnresolved + iPos, iLen );
            *( pSub->pszBefore + iLen ) = '\0';
        }

        /* rewriting and splitting the tag to be replaced */
        gmlResolverSplitRewrite( resolver->pszUnresolved, pSub->iOffsetStart, 
                                 pSub->iOffsetEnd, szTagOpen, szTagClose );
        iLen = strlen( szTagOpen );
        if ( iLen > 0 )
        {
            pSub->pszTagOpen = (char *)CPLMalloc( iLen + 1 );
            strcpy ( pSub->pszTagOpen, szTagOpen );
        }
        iLen = strlen( szTagClose );
        if ( iLen > 0 )
        {
            pSub->pszTagClose = (char *)CPLMalloc( iLen + 1 );
            strcpy ( pSub->pszTagClose, szTagClose );
        }
        iPos = pSub->iOffsetEnd + 1;
    }

    /* saving the last GML fragment */
    iLen = strlen ( resolver->pszUnresolved ) - iPos;
    if ( iLen > 0 )
    {
        resolver->pszAfter = (char *)CPLMalloc( iLen + 1 );
        memcpy ( resolver->pszAfter, resolver->pszUnresolved + iPos, iLen );
        *( resolver->pszAfter + iLen ) = '\0';
    }	

    /* allocating the resolved string */
    iOutLen += strlen ( resolver->pszUnresolved );
    resolver->pszResolved = (char *)CPLMalloc( iOutLen + 1 );
    *(resolver->pszResolved) = '\0';
	
    /* reassembling the resolved GML string */
    for ( iSub = 0; iSub < iCount; iSub++ )
    {
        pSub = *( resolver->pArray + iSub ); 
        if ( pSub->pszBefore != NULL )
            strcat ( resolver->pszResolved, pSub->pszBefore );
        if ( pSub->pszTagOpen != NULL )
            strcat ( resolver->pszResolved, pSub->pszTagOpen );
        if ( pSub->pszString != NULL )
            strcat ( resolver->pszResolved, pSub->pszString );
        if ( pSub->pszTagClose != NULL )
            strcat ( resolver->pszResolved, pSub->pszTagClose );
    }
    if ( resolver->pszAfter != NULL )
        strcat ( resolver->pszResolved, resolver->pszAfter );

    return TRUE;
}

static void gmlResolverReset ( struct huge_file_resolver *resolver)
{
/* resetting a resolver struct to initial state */
    struct huge_file_sub_resolver *pNext;
    struct huge_file_sub_resolver *pSub = resolver->pFirst;
    while ( pSub != NULL)
    {
        pNext = pSub->pNext;
        if ( pSub->pszString != NULL )
            CPLFree ( pSub->pszString );
        if ( pSub->pszBefore != NULL )
            CPLFree ( pSub->pszBefore );
        if ( pSub->pszTagOpen != NULL )
            CPLFree ( pSub->pszTagOpen );
        if ( pSub->pszTagClose != NULL )
            CPLFree ( pSub->pszTagClose );
        CPLFree( pSub );
        pSub = pNext;
    }	
    if ( resolver->pszUnresolved != NULL )
        CPLFree ( resolver->pszUnresolved );
    if ( resolver->pszAfter != NULL )
        CPLFree ( resolver->pszAfter );
    if ( resolver->pszResolved != NULL )
        CPLFree ( resolver->pszResolved );
    if ( resolver->pArray != NULL )
        CPLFree ( resolver->pArray );
    resolver->pszUnresolved = NULL;
    resolver->pszAfter = NULL;
    resolver->pszResolved = NULL;
    resolver->pFirst = NULL;
    resolver->pLast = NULL;
    resolver->pArray = NULL;
}

static void gmlGetCleanEdge ( const char *pszGml, int *iCleanStart,
                              int *iCleanEnd )
{
/* identifying a cleaned <Edge> element */
    const char *pIn = pszGml;
    char szTagName[1024];
    int bTagName = FALSE;
    int bTagEnd = FALSE;
    char *pszTagName = szTagName;
    int iOffsetStart = -1;
    int iOffsetEnd = -1;
    int nTags = 0;
    int iTagTopoCurveOpen = -1;
    int iTagTopoCurveClose = -1;
    int iTagDirectedEdgeOpen = -1;
    int iTagDirectedEdgeClose = -1;
    int iStart = -1;
    int iEnd = -1;

    /* default: the whole GML string is valid */
    *iCleanStart = 0;
    *iCleanEnd = strlen( pszGml );

    /* looping on GML Geometry characters */
    while ( *pIn != '\0')
    {
        /* skipping any white character */
        if ( *pIn == ' ' || *pIn == '\t' || *pIn == '\n' || *pIn == '\r')
        {
            /* we are now expecting a TagName here: <*tag* ...> */
            if ( bTagName == TRUE )
            {
                if ( pszTagName != szTagName )
                {
                    bTagName = FALSE;
                    *pszTagName = '\0';
                }
                pIn ++;
                continue;
            }
            pIn ++;
            continue;
        }

        /* a GML tag starts here */
        if ( *pIn == '<')
        {
            pszTagName = szTagName;
            iOffsetStart = pIn - pszGml;
            pIn++;
            bTagName = TRUE;
            bTagEnd = FALSE;
            continue;
        }

        /* closing GML tag marker */
        if ( *pIn == '/')
        {
            *pIn ++;
            bTagEnd = TRUE;
            continue;
        }

        /* a GML tag ends here */
        if ( *pIn == '>')
        {
            if ( bTagName == TRUE )
                *pszTagName = '\0';
            iOffsetEnd = pIn - pszGml;
            if ( EQUAL(szTagName, "TopoCurve") ||
                 EQUAL(szTagName, "gml:TopoCurve") )
            {
                if ( bTagEnd == TRUE )
                    iTagTopoCurveClose = nTags;
                else
                    iTagTopoCurveOpen = nTags;
            }
            if ( EQUAL(szTagName, "directedEdge") || 
                 EQUAL(szTagName, "gml:directedEdge") )
            {
                if ( bTagEnd == TRUE )
                {
                    iTagDirectedEdgeClose = nTags;
                    iEnd = iOffsetStart;
                }
                else
                {
                    iTagDirectedEdgeOpen = nTags;
                    iStart = iOffsetEnd;
                }
            }
            bTagName = FALSE;
            bTagEnd = FALSE;
            nTags ++;
            *pIn ++;
            continue;
        }

        if ( bTagName == TRUE )
            *pszTagName ++ = *pIn;
        pIn ++;
    }

    /* final check for validity */
    if ( iTagTopoCurveOpen == 0 && iTagDirectedEdgeOpen == 1 &&
         iTagTopoCurveClose == nTags - 1 && iTagDirectedEdgeClose == nTags - 2 )
    {
        /* cleaning any trailing or leading white-space */
        int bIsFirst = TRUE;
        for ( int iChar = iStart; iChar < iEnd; iChar++ )
        {
            if ( pszGml[iChar] == '<' && bIsFirst == TRUE )
            {
                *iCleanStart = iChar;
                bIsFirst = FALSE;
            }
            if ( pszGml[iChar] == '>' )
                *iCleanEnd = iChar + 1;
        }	
    }

}

static int gmlHugeFileResolve( struct huge_file_helper *helper )
{
/* identifying any not yet resolved GML string */
    const char * osCommand;
    char *pszErrMsg = NULL;
    int rc;
    int iStartID = 0;
    sqlite3 *hDB = helper->hDB;
    sqlite3_stmt *hQueryStmt;
    sqlite3_stmt *hUpdateStmt;
    sqlite3_stmt *hExploredStmt;
    struct huge_file_resolver resolver;
    struct huge_file_sub_resolver *pSub;
    int iOutCleanStart;
    int iOutCleanEnd;	

    /* initialing an empty resolver struct */
    resolver.iRowID = -1;
    resolver.iItemID = -1;
    resolver.pszUnresolved = NULL;
    resolver.pszAfter = NULL;
    resolver.pszResolved = NULL;
    resolver.pFirst = NULL;
    resolver.pLast = NULL;
    resolver.pArray = NULL;

    /* query cursor */
    osCommand = "SELECT s.ROWID, s.gml_string, h.offset_start, "
                "h.offset_end, s2.gml_resolved, h.item_id, "
                "s2.clean_start, s2.clean_end "
                "FROM gml_strings AS s "
                "JOIN gml_hrefs AS h ON (s.item_id = h.item_id) "
                "JOIN gml_strings AS s2 ON (s2.gml_id = h.gml_href) "
                "WHERE s.gml_resolved IS NULL AND h.explored IS NULL "
                "AND s.item_id BETWEEN ? AND ?";
    rc = sqlite3_prepare( hDB, osCommand, -1, &hQueryStmt, NULL );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to create QUERY stmt for resolver" );
        return FALSE;
    }

    /* update cursor */
    osCommand = "UPDATE gml_strings "
                "SET gml_resolved = ?, "
                "clean_start = ?, "
                "clean_end = ?, "
                "gml_string = NULL "
                "WHERE ROWID = ?";
    rc = sqlite3_prepare( hDB, osCommand, -1, &hUpdateStmt, NULL );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to create UPDATE stmt for resolved" );
        sqlite3_finalize ( hQueryStmt );
        return FALSE;
    }

    /* update cursor (explored) */
    osCommand = "UPDATE gml_hrefs "
                "SET explored = 1 "
                "WHERE item_id = ?";
    rc = sqlite3_prepare( hDB, osCommand, -1, &hExploredStmt, NULL );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to create UPDATE stmt for explored" );
        sqlite3_finalize ( hQueryStmt );
        sqlite3_finalize ( hUpdateStmt );
        return FALSE;
    }

    int bOkLoop = TRUE;
    while ( bOkLoop == TRUE )
    {
        bOkLoop = FALSE;

        /* starting a TRANSACTION */
        rc = sqlite3_exec( hDB, "BEGIN", NULL, NULL, &pszErrMsg );
        if( rc != SQLITE_OK )
        {
            CPLError( CE_Failure, CPLE_AppDefined, 
                      "Unable to perform BEGIN TRANSACTION: %s",
                      pszErrMsg );
            sqlite3_free( pszErrMsg );
            sqlite3_finalize ( hQueryStmt );
            sqlite3_finalize ( hUpdateStmt );
            sqlite3_finalize ( hExploredStmt );
            return FALSE;
        }
        sqlite3_reset ( hQueryStmt );
        if ( iStartID < helper->iItemId )
        {
            sqlite3_bind_int ( hQueryStmt, 1, iStartID );
            sqlite3_bind_int ( hQueryStmt, 2, iStartID + 5000 );
            iStartID += 5000;
            bOkLoop = TRUE;
        }
        else
        {
            sqlite3_bind_int ( hQueryStmt, 1, 0 );
            sqlite3_bind_int ( hQueryStmt, 2, helper->iItemId + 1 );
        }
	
        /* looping on the QUERY result-set */
        while ( TRUE )
        {
            int iRowID;
            const char *pszString;
            int iOffsetStart;
            int iOffsetEnd;
            const char *pszResolved;
            int iItemID;
            int iCleanStart;
            int iCleanEnd;				
            int iLen;

            rc = sqlite3_step( hQueryStmt );
            if ( rc == SQLITE_ROW )
            {
                iRowID = sqlite3_column_int( hQueryStmt, 0 );
                pszString = (const char *) sqlite3_column_blob( hQueryStmt, 1 );
                iOffsetStart = sqlite3_column_int( hQueryStmt, 2 );
                iOffsetEnd = sqlite3_column_int( hQueryStmt, 3 );
                if ( sqlite3_column_type( hQueryStmt, 4 ) == SQLITE_NULL )
                    pszResolved = NULL;
                else
                    pszResolved = (const char *)sqlite3_column_blob(hQueryStmt, 4);
                iItemID = sqlite3_column_int( hQueryStmt, 5 );
                iCleanStart = sqlite3_column_int( hQueryStmt, 6 );
                iCleanEnd = sqlite3_column_int( hQueryStmt, 7 );

                /* item changed: attempting to resolve */
                if ( iRowID != resolver.iRowID )
                {
                    if ( gmlResolverResolve ( &resolver ) == TRUE )
                    {
                        /* updating any resolved GML string */
                        sqlite3_reset ( hUpdateStmt );
                        sqlite3_clear_bindings ( hUpdateStmt );
                        sqlite3_bind_blob( hUpdateStmt, 1,
                                           resolver.pszResolved, 
                                           strlen(resolver.pszResolved), 
                                          SQLITE_STATIC );
                        gmlGetCleanEdge ( resolver.pszResolved, 
                                          &iOutCleanStart, &iOutCleanEnd ); 
                        sqlite3_bind_int (hUpdateStmt, 2, iOutCleanStart);
                        sqlite3_bind_int (hUpdateStmt, 3, iOutCleanEnd);
                        sqlite3_bind_int ( hUpdateStmt, 4, resolver.iRowID );
                        rc = sqlite3_step( hUpdateStmt );
                        if( rc != SQLITE_OK && rc != SQLITE_DONE )
                        {
                            CPLError( CE_Failure, CPLE_AppDefined,
                                      "sqlite3_step() failed:\n  %s",
                                      sqlite3_errmsg(hDB) );
                            sqlite3_finalize ( hQueryStmt );
                            sqlite3_finalize ( hUpdateStmt );
                            sqlite3_finalize ( hExploredStmt );
                            return FALSE;
                        }
                        bOkLoop = TRUE;

                        /* marking EXPLORED references */
                        sqlite3_reset ( hExploredStmt );
                        sqlite3_clear_bindings ( hExploredStmt );
                        sqlite3_bind_int ( hExploredStmt, 1, resolver.iItemID );
                        rc = sqlite3_step( hExploredStmt );
                        if( rc != SQLITE_OK && rc != SQLITE_DONE )
                        {
                            CPLError( CE_Failure, CPLE_AppDefined,
                                      "sqlite3_step() failed:\n  %s",
                                      sqlite3_errmsg(hDB) );
                            sqlite3_finalize ( hQueryStmt );
                            sqlite3_finalize ( hUpdateStmt );
                            sqlite3_finalize ( hExploredStmt );
                            return FALSE;
                        }
                    }

                    gmlResolverReset( &resolver );
                    resolver.iRowID = iRowID;
                    iLen = strlen( pszString );
                    resolver.pszUnresolved = (char *)CPLMalloc( iLen + 1 );
                    strcpy ( resolver.pszUnresolved, pszString );
                    resolver.iItemID = iItemID;
                }

                /* appending a sub item into the resolver */
                pSub = (struct huge_file_sub_resolver *) CPLMalloc( 
                        sizeof( struct huge_file_sub_resolver ) );

                /* setting up the sub item */
                if ( pszResolved == NULL )
                    pSub->pszString = NULL;
                else
                {
                    iLen = iCleanEnd - iCleanStart;
                    if ( iLen > 0)
                    {
                        pSub->pszString = (char *)CPLMalloc ( iLen + 1 );
                        memcpy( pSub->pszString, pszResolved + iCleanStart, 
                                iLen );
                        *( pSub->pszString + iLen ) = '\0';
                    }
                    else
                        pSub->pszString = NULL;
                }
                pSub->pszBefore = NULL;
                pSub->pszTagOpen = NULL;
                pSub->pszTagClose = NULL;
                pSub->iOffsetStart = iOffsetStart;
                pSub->iOffsetEnd = iOffsetEnd;
                pSub->pNext = NULL;

                /* inserting the sub item into the linked list */
                if ( resolver.pFirst == NULL )
                    resolver.pFirst = pSub;
                if ( resolver.pLast != NULL )
                    resolver.pLast->pNext = pSub;
                resolver.pLast = pSub;
            }
            else
                break;
        }

	/* attempting to resolve the last item */
        if ( gmlResolverResolve ( &resolver ) == TRUE )
        {
            /* updating any resolved GML string */
            sqlite3_reset ( hUpdateStmt );
            sqlite3_clear_bindings ( hUpdateStmt );
            sqlite3_bind_blob( hUpdateStmt, 1, resolver.pszResolved, 
                               strlen(resolver.pszResolved), SQLITE_STATIC );
            gmlGetCleanEdge ( resolver.pszResolved, &iOutCleanStart, 
                              &iOutCleanEnd ); 
            sqlite3_bind_int (hUpdateStmt, 2, iOutCleanStart);
            sqlite3_bind_int (hUpdateStmt, 3, iOutCleanEnd);
            sqlite3_bind_int ( hUpdateStmt, 4, resolver.iRowID );
            rc = sqlite3_step( hUpdateStmt );
            if( rc != SQLITE_OK && rc != SQLITE_DONE )
            {
                CPLError( CE_Failure, CPLE_AppDefined,
                          "sqlite3_step() failed:\n  %s",
                          sqlite3_errmsg(hDB) );
                sqlite3_finalize ( hQueryStmt );
                sqlite3_finalize ( hUpdateStmt );
                sqlite3_finalize ( hExploredStmt );
                return FALSE;
            }
            bOkLoop = TRUE;

            /* marking EXPLORED references */
            sqlite3_reset ( hExploredStmt );
            sqlite3_clear_bindings ( hExploredStmt );
            sqlite3_bind_int ( hExploredStmt, 1, resolver.iItemID );
            rc = sqlite3_step( hExploredStmt );
            if( rc != SQLITE_OK && rc != SQLITE_DONE )
            {
                CPLError( CE_Failure, CPLE_AppDefined,
                          "sqlite3_step() failed:\n  %s",
                          sqlite3_errmsg(hDB) );
                sqlite3_finalize ( hQueryStmt );
                sqlite3_finalize ( hUpdateStmt );
                sqlite3_finalize ( hExploredStmt );
                return FALSE;
            }
        }
        gmlResolverReset( &resolver );

        /* confirming the still pending TRANSACTION */
        rc = sqlite3_exec( hDB, "COMMIT", NULL, NULL, &pszErrMsg );
        if( rc != SQLITE_OK )
        {
            CPLError( CE_Failure, CPLE_AppDefined, 
                      "Unable to perform COMMIT TRANSACTION: %s",
                      pszErrMsg );
            sqlite3_free( pszErrMsg );
            return FALSE;
        }		
    }
    sqlite3_finalize ( hQueryStmt );
    sqlite3_finalize ( hUpdateStmt );
    sqlite3_finalize ( hExploredStmt );
    return TRUE;
}
	
static void gmlHugeFileEvalStart( struct huge_file_helper *helper,
                                  const char *pszTagName, 
                                  const char *pszTagKey,
                                  const char *pszTagValue, 
                                  int iOffsetStart )
{
/* 
/ this function evalutates if <gml:id> or <xlink:href> are defined
/ eventually updating the internal helper struct as appropriate
*/
    struct huge_file_tag_helper *pItem = NULL;

    /* testing for <directedNode orientation="-"> */
    if ( EQUAL(pszTagName, "directedNode") == TRUE || 
         EQUAL(pszTagName, "gml:directedNode" ) == TRUE )
    {
        if ( EQUAL(pszTagKey, "orientation") == TRUE && 
             EQUAL(pszTagValue, "-") == TRUE )
            helper->bFromNode = TRUE;
    }

    /* testing <LineString srsName="...">  or <Curve srsName="..."> */
    if ( EQUAL(pszTagName, "LineString") == TRUE || 
         EQUAL(pszTagName, "gml:LineString" ) == TRUE ||
         EQUAL(pszTagName, "Curve") == TRUE || 
         EQUAL(pszTagName, "gml:Curve" ) == TRUE )
    {
        if ( EQUAL(pszTagKey, "srsName") == TRUE)
            strcpy ( helper->szNodeSrs, pszTagValue );
    }

    if ( EQUAL(pszTagKey, "gml:id") == TRUE || 
         EQUAL(pszTagKey, "xlink:href") == TRUE )
        ;
    else
        return;

    /* appending a new tag item to the list */
    pItem = (struct huge_file_tag_helper *) CPLMalloc( 
            sizeof(struct huge_file_tag_helper) );
    if ( pItem == NULL )
        return;
		
    /* initializing the new tag item */
    *( pItem->szTagName ) = '\0';
    *( pItem->szIdHref ) = '\0';
    *( pItem->szNodeFrom ) = '\0';
    *( pItem->szNodeTo ) = '\0';
    pItem->iOffsetStart = -1;
    pItem->iOffsetEnd = -1;
    pItem->bHasCoords = FALSE;
    pItem->bHasZ = FALSE;
    pItem->bIsNodeGmlId = FALSE;
    if ( EQUAL(pszTagName, "Node") == TRUE || 
         EQUAL(pszTagName, "gml:Node" ) == TRUE )
        pItem->bIsNodeGmlId = TRUE;
    pItem->pPrev = helper->pLast;
    pItem->pNext = NULL;

    /* appending the item to the linked list */
    if ( helper->pFirst == NULL )
        helper->pFirst = pItem;
    if ( helper->pLast != NULL )
        helper->pLast->pNext = pItem;
    helper->pLast = pItem;

    /* setting tag values */
    strcpy ( pItem->szTagName, pszTagName );
    pItem->iOffsetStart = iOffsetStart;

    /* testing for <gml:id> */
    if ( EQUAL(pszTagKey, "gml:id") == TRUE )
    {
         strcpy ( pItem->szIdHref, pszTagValue );
         pItem->bIsGmlHref = FALSE;
    }

    /* testing for <xlink:href> */
    if ( EQUAL(pszTagKey, "xlink:href") == TRUE )
    {
         strcpy ( pItem->szIdHref, pszTagValue + 1 );
         pItem->bIsGmlHref = TRUE;
    }
}

static void gmlHugeFileEvalEnd( struct huge_file_helper *helper, 
                                const char *pszTagName, int iOffsetEnd )
{
/* 
/ this function attempts to update the internal helper struct
/ when a GML closing tag is found
*/
    bool bFromNode = helper->bFromNode;
    struct huge_file_tag_helper *pItem = helper->pLast;

    if ( EQUAL(pszTagName, "directedNode") == TRUE || 
         EQUAL(pszTagName, "gml:directedNode" ) == TRUE )
        helper->bFromNode = FALSE;

    /* scanning the list in reverse direction */
    while ( pItem != NULL)
    {
        if ( EQUAL(pszTagName, pItem->szTagName) == TRUE && 
             pItem->iOffsetEnd < 0 )
        {
            pItem->iOffsetEnd = iOffsetEnd;

            /* saving Node IDs */
            if ( EQUAL(pItem->szTagName, "Node") == TRUE || 
                 EQUAL(pItem->szTagName, "gml:Node" ) == TRUE ||
                 EQUAL(pItem->szTagName, "directedNode") == TRUE || 
                 EQUAL(pItem->szTagName, "gml:directedNode" ) == TRUE )
            {
                if ( bFromNode == TRUE )
                    strcpy ( helper->szNodeFrom, pItem->szIdHref );
                else
                    strcpy ( helper->szNodeTo, pItem->szIdHref );
            }

            /* assigning Node IDs to Edge */
            if ( EQUAL(pItem->szTagName, "Edge") == TRUE || 
                 EQUAL(pItem->szTagName, "gml:Edge" ) == TRUE )
            {
                strcpy ( pItem->szNodeFrom, helper->szNodeFrom );
                strcpy ( pItem->szNodeTo, helper->szNodeTo );
            }
            return;
        }
        pItem = pItem->pPrev;
    }
}

static void gmlHugeFileNodeCoords( struct huge_file_helper *helper, 
                                   const char *pszGml )
{
/* 
/ this function attempts to set coordinates for <Node> items
/ when required (a LINESTRING is expected to be processed)
*/
    double x0;
    double y0;
    double z0;
    double x1;
    double y1;
    double z1;
    int bHasZ = FALSE;
    OGRLineString* poLineString = NULL;
    struct huge_file_tag_helper *pItem;

    /* looping on GML tags */
    void* hCacheSRS = GML_BuildOGRGeometryFromList_CreateCache();    
    pItem = helper->pFirst;
    while ( pItem != NULL )
    {
        /* only applies to <Egde> elements */
        if ( EQUAL(pItem->szTagName, "Edge") == TRUE || 
             EQUAL(pItem->szTagName, "gml:Edge") == TRUE )
            ;
        else
        {
            pItem = pItem->pNext;
            continue;
        }

        CPLXMLNode * papsNode = CPLParseXMLString(pszGml);
        CPLXMLNode * papsGeometry[2];
        papsGeometry[0] = papsNode;
        papsGeometry[1] = NULL;
        OGRGeometry *poGeometry = GML_BuildOGRGeometryFromList( papsGeometry, 
                                                                FALSE, FALSE, 
                                                                NULL, FALSE, 
                                                                FALSE, hCacheSRS);
        CPLDestroyXMLNode(papsNode);
        if ( poGeometry == NULL )
        {
            pItem = pItem->pNext;
            continue;
        }

        int gType = wkbFlatten(poGeometry->getGeometryType());
        if ( gType == wkbLineString || gType == wkbLinearRing )
            poLineString = (OGRLineString*) poGeometry;
        else if ( gType == wkbMultiLineString )
        {
            OGRGeometryCollection* poGeomCollection = 
                (OGRGeometryCollection*) poGeometry;
            if ( poGeomCollection->getNumGeometries() == 1 )
            {
                gType = wkbFlatten(poGeomCollection->getGeometryRef(0)->getGeometryType());
                if ( gType == wkbLineString || gType == wkbLinearRing )
                    poLineString = (OGRLineString*) poGeomCollection->getGeometryRef(0);
            }
        }
        if ( poLineString != NULL )
        {		
            x0 = poLineString->getX(0);
            y0 = poLineString->getY(0);
            if ( poGeometry->getCoordinateDimension() == 3 )
            {
                z0 = poLineString->getZ(0);
                bHasZ = TRUE;
            }
            x1 = poLineString->getX( poLineString->getNumPoints() - 1 );
            y1 = poLineString->getY( poLineString->getNumPoints() - 1 );
            if ( poGeometry->getCoordinateDimension() == 3 )
            z1 = poLineString->getZ( poLineString->getNumPoints() - 1 );

            /* assigning coords to each <Node> */
            pItem->bHasCoords = TRUE;
            pItem->bHasZ = bHasZ;
            pItem->xNodeFrom = x0;
            pItem->yNodeFrom = y0;
            if ( bHasZ == TRUE )
                pItem->zNodeFrom = z0;
            pItem->xNodeTo = x1;
            pItem->yNodeTo = y1;
            if ( bHasZ == TRUE )
                pItem->zNodeTo = z1;
        }		
        delete poGeometry;
        pItem = pItem->pNext;
    }
    GML_BuildOGRGeometryFromList_DestroyCache(hCacheSRS);
}

static int gmlHugeFileSQLiteInsert( struct huge_file_helper *helper,
                                    const char *pszGml )
{
/* inserting any appropriate row into the SQLite DB */
    int rc;
    int bGmlIdDone = FALSE;
    struct huge_file_tag_helper *pItem;

    /* looping on GML tags */
    pItem = helper->pFirst;
    while ( pItem != NULL )
    {
       /* <Node> <Edge> reference found */
       if ( pItem->bHasCoords == TRUE )
       {
            sqlite3_reset ( helper->hNodes );
            sqlite3_clear_bindings ( helper->hNodes );
            sqlite3_bind_text( helper->hNodes, 1, pItem->szNodeFrom, -1, 
                               SQLITE_STATIC );
            sqlite3_bind_double ( helper->hNodes, 2, pItem->xNodeFrom );
            sqlite3_bind_double ( helper->hNodes, 3, pItem->yNodeFrom );
            if ( pItem->bHasZ == TRUE )
                sqlite3_bind_double ( helper->hNodes, 4, pItem->zNodeFrom );
            if ( *(helper->szNodeSrs) > 0 )
                sqlite3_bind_text ( helper->hNodes, 5, helper->szNodeSrs, -1,
                                    SQLITE_STATIC );
            rc = sqlite3_step( helper->hNodes );
            if( rc != SQLITE_OK && rc != SQLITE_DONE )
            {
                CPLError( CE_Failure, CPLE_AppDefined, 
                          "sqlite3_step() failed:\n  %s", 
                          sqlite3_errmsg(helper->hDB) );
                return FALSE;
            }
            sqlite3_reset ( helper->hNodes );
            sqlite3_clear_bindings ( helper->hNodes );
            sqlite3_bind_text( helper->hNodes, 1, pItem->szNodeTo, -1, 
                               SQLITE_STATIC );
            sqlite3_bind_double ( helper->hNodes, 2, pItem->xNodeTo );
            sqlite3_bind_double ( helper->hNodes, 3, pItem->yNodeTo );
            if ( pItem->bHasZ == TRUE )
                sqlite3_bind_double ( helper->hNodes, 4, pItem->zNodeTo );
            if ( *(helper->szNodeSrs) > 0 )
                sqlite3_bind_text ( helper->hNodes, 5, helper->szNodeSrs, -1,
                                    SQLITE_STATIC );
            rc = sqlite3_step( helper->hNodes );
            if( rc != SQLITE_OK && rc != SQLITE_DONE )
            {
                CPLError( CE_Failure, CPLE_AppDefined, 
                          "sqlite3_step() failed:\n  %s", 
                sqlite3_errmsg(helper->hDB) );
                return FALSE;
            }
        }

        /* <gml:id> and <xlink:href> */
        if ( strlen( pItem->szIdHref ) > 0 )
        {
            /* <gml:id> */
            if ( pItem->bIsGmlHref == FALSE && 
                 ( bGmlIdDone == FALSE || pItem->bIsNodeGmlId == TRUE ) )
            {
                int iLen;
                char *pszStr;
                if ( pItem->bIsNodeGmlId == TRUE )
                {
                    iLen = ( pItem->iOffsetEnd - pItem->iOffsetStart ) + 1;
                    pszStr = (char *) CPLMalloc(iLen + 1);
                    memcpy(pszStr, pszGml + pItem->iOffsetStart, iLen);
                    pszStr[iLen] = '\0';
                }
                else
                {
                    iLen = strlen(pszGml);
                    pszStr = (char *) CPLMalloc(iLen + 1);
                    strcpy( pszStr, pszGml );
                }
                sqlite3_reset ( helper->hStrings );
                sqlite3_clear_bindings ( helper->hStrings );
                sqlite3_bind_int ( helper->hStrings, 1, helper->iItemId );
                sqlite3_bind_text( helper->hStrings, 2, pItem->szIdHref, -1,
                                   SQLITE_STATIC );
                if ( strstr ( pszStr, "xlink:href" ) == NULL )
                {
                    int iCleanStart;
                    int iCleanEnd;
                    gmlGetCleanEdge ( pszStr, &iCleanStart, &iCleanEnd ); 
                    sqlite3_bind_null (helper->hStrings, 3);
                    sqlite3_bind_blob( helper->hStrings, 4, pszStr, iLen, 
                                       CPLFree );
                    sqlite3_bind_int (helper->hStrings, 5, iCleanStart);
                    sqlite3_bind_int (helper->hStrings, 6, iCleanEnd);
                }
                else
                {
                    sqlite3_bind_blob( helper->hStrings, 3, pszStr, iLen, 
                                       CPLFree );
                    sqlite3_bind_null (helper->hStrings, 4);
                }
                rc = sqlite3_step( helper->hStrings );
                if( rc != SQLITE_OK && rc != SQLITE_DONE )
                {
                    CPLError( CE_Failure, CPLE_AppDefined, 
                              "sqlite3_step() failed:\n  %s", 
                              sqlite3_errmsg(helper->hDB) );
                    return FALSE;
                }
                if (pItem->bIsNodeGmlId == FALSE )
                    bGmlIdDone = TRUE;
            }

            /* <xlink:href> (for cross reference validation) */
            else
            {
                sqlite3_reset ( helper->hHRefs );
                sqlite3_clear_bindings ( helper->hHRefs );
                sqlite3_bind_int ( helper->hHRefs, 1, helper->iItemId );
                sqlite3_bind_text ( helper->hHRefs, 2, pItem->szIdHref, -1,
                                    SQLITE_STATIC );
                sqlite3_bind_int ( helper->hHRefs, 3, pItem->iOffsetStart );
                sqlite3_bind_int ( helper->hHRefs, 4, pItem->iOffsetEnd );
                rc = sqlite3_step( helper->hHRefs );
                if( rc != SQLITE_OK && rc != SQLITE_DONE )
                {
                    CPLError( CE_Failure, CPLE_AppDefined, 
                              "sqlite3_step() failed:\n  %s", 
                              sqlite3_errmsg(helper->hDB) );
                    return FALSE;
                }
            }
        }
        pItem = pItem->pNext;
    }
    return TRUE;
}

static void gmlHugeFileReset( struct huge_file_helper *helper )
{
/* resetting an empty helper struct */
    struct huge_file_tag_helper *pNext;
    struct huge_file_tag_helper *p = helper->pFirst;

    /* cleaning any previous item */
    while ( p != NULL )
    {
        pNext = p->pNext;
        CPLFree ( p );
        p = pNext;
    }
    helper->pFirst = NULL;
    helper->pLast = NULL;
    helper->bFromNode = FALSE;
    *( helper->szNodeFrom ) = '\0';
    *( helper->szNodeTo ) = '\0';
    *( helper->szNodeSrs ) = '\0';
}

static int gmlHugeFileCheckXrefs( struct huge_file_helper *helper, 
                                  const char *pszGml )
{
/*
/ this function will parse a single GML fragment representing
/ some Geometry [this being the result of previous Expat/Xerces
/ parsing, so  we can safely assume it to represent clean and 
/ well-formatted GML]
/
/ the scope of this function is the one to identify any
/ <gml:id="xxx"> or <xlink:href="xxx"> value, thus updating
/ the SQLite DB-file.
/
/ a small complication arise because <Edge>, <directedEdge>, ...
/ tag can contain nested <Node> items
*/
    char szTagName[1024];
    char szTagKey[1024];
    char szTagValue[1024];
    char *pszTagName = szTagName;
    char *pszTagKey = szTagKey;
    char *pszTagValue = szTagValue;
    int iOffsetStart = -1;
    int iOffsetEnd = -1;
    int bTagEnd = FALSE;
    int bTagName = FALSE;
    int bTagKey = FALSE;
    int bTagValue = FALSE;
    int bString = FALSE;
    const char *pIn = pszGml;

    /* resetting an empty helper struct */
    gmlHugeFileReset ( helper );

    /* looping on GML Geometry characters */
    while ( *pIn != '\0')
    {
        /* 
        / we already have a TagName and KeyName set
        / so we are expecting a KeyValue here: <tag key=*value*>
        */
        if ( bTagValue == TRUE)
        {
            /* opening / closing string value */
            if ( *pIn == '"')
            {
                /* string starts here */
                if ( bString == FALSE )
                {
                    bString = TRUE;
                    pszTagValue = szTagValue;
                    pIn ++;
                    continue;
                }

                /* string ends here */
                else
                {
                    *pszTagValue = '\0';
                    gmlHugeFileEvalStart( helper, szTagName, szTagKey, 
                                          szTagValue, iOffsetStart );
                    bString = FALSE;
                    bTagValue = FALSE;
                    bTagKey = TRUE;
                    pszTagKey = szTagKey;
                    pIn ++;
                    continue;
                }
            }
            if ( bString == TRUE )
                *pszTagValue++ = *pIn;
            pIn ++;
            continue;
        }

        /* skipping any white character */
        if ( *pIn == ' ' || *pIn == '\t' || *pIn == '\n' || *pIn == '\r')
        {
            /* we are now expecting a TagName here: <*tag* ...> */
            if ( bTagName == TRUE )
            {
                if ( pszTagName != szTagName )
                {
                    bTagName = FALSE;
                    *pszTagName = '\0';
                    bTagKey = TRUE;
                    pszTagKey = szTagKey;
                }
                pIn ++;
                continue;
            }

            /* we are now expecting a KeyName here: <tag *key*=...> */
            if ( bTagKey == TRUE )
            {
                if ( pszTagKey != szTagKey )
                {
                    bTagKey = FALSE;
                    *pszTagKey = '\0';
                    bTagValue = TRUE;
                    bString = FALSE;
                }
                pIn ++;
                continue;
            }
            pIn ++;
            continue;
        }

        /* Key=Value separator */
        if ( *pIn == '=')
        {
            /* we are now expecting a KeyValue here: <tag key *=* ...> */
            if ( bTagKey == TRUE )
            {
                bTagKey = FALSE;
                *pszTagKey = '\0';
                bTagValue = TRUE;
                bString = FALSE;
            }
            pIn ++;
            continue;
        }

        /* a GML tag starts here */
        if ( *pIn == '<')
        {
            pszTagName = szTagName;
            iOffsetStart = pIn - pszGml;
            pIn++;
            bTagName = TRUE;
            bTagKey = FALSE;
            bTagValue = FALSE;
            bTagEnd = FALSE;
            continue;
        }

        /* closing GML tag marker */
        if ( *pIn == '/')
        {
            *pIn ++;
            bTagEnd = TRUE;
            continue;
        }

        /* a GML tag ends here */
        if ( *pIn == '>')
        {
            if ( bTagName == TRUE )
                *pszTagName = '\0';
            iOffsetEnd = pIn - pszGml;
            if ( bTagEnd == TRUE )
                gmlHugeFileEvalEnd( helper, szTagName, iOffsetEnd );
            if ( bTagEnd == TRUE &&  
                 ( EQUAL(szTagName, "Edge") == TRUE || 
                   EQUAL(pszTagName, "gml:Edge" ) == TRUE ) )
            {
                helper->bFromNode = FALSE;
                *( helper->szNodeFrom ) = '\0';
                *( helper->szNodeTo ) = '\0';
            }
            bTagName = FALSE;
            bTagKey = FALSE;
            bTagValue = FALSE;
            bTagEnd = FALSE;
            *pIn ++;
            continue;
        }

        if ( bTagName == TRUE )
        *pszTagName ++ = *pIn;
        if ( bTagKey == TRUE )
            *pszTagKey ++ = *pIn;
        pIn ++;
    }

    /* retrieving Node coordinates when required */
    gmlHugeFileNodeCoords( helper, pszGml );

    /* and finally inserting into the SQLite DB any appropriate row */
    return gmlHugeFileSQLiteInsert ( helper, pszGml );
}

static void gmlHugeFileCleanUp ( struct huge_file_helper *helper ) 
{
/* cleaning up any SQLite handle */
    if ( helper->hNodes != NULL )
        sqlite3_finalize ( helper->hNodes );
    if ( helper->hStrings != NULL )
        sqlite3_finalize ( helper->hStrings );
    if ( helper->hHRefs != NULL )
        sqlite3_finalize ( helper->hHRefs );
    if ( helper->hDB != NULL )
        sqlite3_close( helper->hDB );
}

static void gmlUpdateFeatureClasses ( struct huge_class_container * pCC, 
                                      GMLReader *pReader,
                                      int *m_bSequentialLayers )
{
/* updating the FeatureClass list */
    int clIdx;
    for (clIdx = 0; clIdx < pReader->GetClassCount(); clIdx++)
    {
        GMLFeatureClass* poClass = pReader->GetClass( clIdx );
        if (poClass != NULL)
            poClass->SetFeatureCount( 0 ); 
    }
    int m_bValid = FALSE;
    struct huge_class_item *pCI = pCC->pFirst;
    while ( pCI != NULL )
    { 
    /* updating Classes */
        GMLFeatureClass* poClass = pReader->GetClass( pCI->m_pszName );
        if (poClass != NULL)
        {
            poClass->SetFeatureCount( pCI->n_nItemCount );
            if ( pCI->n_nGeomCount == 0)
                poClass->SetGeometryType( wkbNone );
            m_bValid = TRUE;
        }
        pCI = pCI->pNext;
    }
    if ( m_bValid == TRUE && pCC->m_bSequentialLayers == TRUE )
        *m_bSequentialLayers = TRUE;
}
	
static int gmlHugeFileWriteResolved ( struct huge_file_helper *helper, 
                                      const char *pszOutputFilename, 
                                      GMLReader *pReader,
                                      int *m_bSequentialLayers ) 
{
/* writing the resolved GML file */
    VSILFILE       *fp;
    GMLFeature  *poFeature;
    int iFeatureUID = 0;
    const char * osCommand;
    int rc;
    sqlite3 *hDB = helper->hDB;
    sqlite3_stmt *hStmt;
    sqlite3_stmt *hStmtNodes;

/* -------------------------------------------------------------------- */
/*      Opening the resolved GML file                                   */
/* -------------------------------------------------------------------- */
    fp = VSIFOpenL( pszOutputFilename, "w" );
    if( fp == NULL )
    {
        CPLError( CE_Failure, CPLE_OpenFailed, 
                  "Failed to open %.500s to write.", pszOutputFilename );
        return FALSE;
    }

/* query cursor */
    osCommand = "SELECT gml_resolved "
                "FROM gml_strings "
                "WHERE item_id = ?";
    rc = sqlite3_prepare( hDB, osCommand, -1, &hStmt, NULL );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to create QUERY stmt for rewrite resolved" );
        VSIFCloseL( fp );
        return FALSE;
    }
    /* query cursor [Nodes] */
    osCommand = "SELECT gml_id, x, y, z, srs "
                "FROM nodes";
    rc = sqlite3_prepare( hDB, osCommand, -1, &hStmtNodes, NULL );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to create QUERY stmt for NODES" );
        VSIFCloseL( fp );
        sqlite3_finalize ( hStmt );
        return FALSE;
    }


    VSIFPrintfL ( fp, "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n" ); 
    VSIFPrintfL ( fp, "<ResolvedTopoFeatureCollection  "
                      "xmlns:gml=\"http://www.opengis.net/gml\">\n" );		
    VSIFPrintfL ( fp, "  <ResolvedTopoFeatureMembers>\n" );
    /* exporting Nodes */
    struct huge_class_container * pCC = gmlAllocClassContainer();
    while ( TRUE )
    {
        rc = sqlite3_step( hStmtNodes );
        if ( rc == SQLITE_ROW )
        {
            const char *pszGmlId;
            double x;
            double y;
            double z = 0.0;
            int bHasZ = FALSE;
            const char *pszSrs = NULL;
            pszGmlId = (const char *)sqlite3_column_text ( hStmtNodes, 0 );
            x = sqlite3_column_double ( hStmtNodes, 1 );
            y = sqlite3_column_double ( hStmtNodes, 2 );
            if ( sqlite3_column_type( hStmtNodes, 3 ) == SQLITE_FLOAT )
            {
                z = sqlite3_column_double ( hStmtNodes, 3 );
                bHasZ = TRUE;
            }
            if ( sqlite3_column_type (hStmtNodes, 4 ) == SQLITE_TEXT )
                pszSrs = (const char *)sqlite3_column_text ( hStmtNodes, 4 );

            /* inserting a node into the resolved GML file */
			gmlUpdateClass( pCC, "ResolvedNodes", TRUE );
            VSIFPrintfL ( fp, "    <ResolvedNodes>\n" );
            VSIFPrintfL ( fp, "      <NodeGmlId>\%s</NodeGmlId>\n", pszGmlId );
            VSIFPrintfL ( fp, "      <ResolvedGeometry> \n" );
            if ( pszSrs == NULL )
                VSIFPrintfL ( fp, "        <gml:Point srsDimension=\"%d\">",
                              ( bHasZ == TRUE ) ? 3 : 2 );
            else
                VSIFPrintfL ( fp, "        <gml:Point srsDimension=\"%d\""
                                  " srsName=\"%s\">",
                              ( bHasZ == TRUE ) ? 3 : 2, pszSrs );
            if ( bHasZ == TRUE )
                VSIFPrintfL ( fp, "<gml:pos>%1.8f %1.8f %1.8f</gml:pos>"
                                  "</gml:Point>\n",
                              x, y, z );
            else
                VSIFPrintfL ( fp, "<gml:pos>%1.8f %1.8f</gml:pos>"
                                  "</gml:Point>\n",
                              x, y );
            VSIFPrintfL ( fp, "      </ResolvedGeometry> \n" );
            VSIFPrintfL ( fp, "    </ResolvedNodes>\n" );
        }
        else
            break;
    }

    /* processing GML features */
    helper->iItemId = 0;
    while( (poFeature = pReader->NextFeature()) != NULL )
    {
        GMLFeatureClass *poClass = poFeature->GetClass();
        const CPLXMLNode* const * papszGeomList = poFeature->GetGeometryList();
        int iPropCount = poClass->GetPropertyCount();
		
        int b_has_geom = FALSE;
        VSIFPrintfL ( fp, "    <%s>\n", poClass->GetElementName() ); 

        for ( int iProp = 0; iProp < iPropCount; iProp++ )
        {
            GMLPropertyDefn *poPropDefn = poClass->GetProperty( iProp );
            const char *pszPropName = poPropDefn->GetName();
            const GMLProperty *poProp = poFeature->GetProperty( iProp );
            for ( int iSub = 0; iSub < poProp->nSubProperties; iSub++ )
            {
                VSIFPrintfL ( fp, "      <%s>\%s</%s>\n", 
                              pszPropName, 
                              poProp->papszSubProperties[iSub], 
                              pszPropName ); 
            }
        }
		
        if( papszGeomList != NULL )
        {
            int i = 0;
			const CPLXMLNode *pszNode = papszGeomList[i];
            while(pszNode)
            {
                bool bToBeFreed = FALSE;
				char *pszResolved = NULL;

                VSIFPrintfL ( fp, "      <ResolvedGeometry> \n" );

                sqlite3_reset ( hStmt );
                sqlite3_bind_int ( hStmt, 1, helper->iItemId );
                rc = sqlite3_step( hStmt );
                if ( rc == SQLITE_ROW )
                {
                    if ( sqlite3_column_type( hStmt, 0 ) != SQLITE_NULL )
                        pszResolved = (char *) sqlite3_column_blob( hStmt,
                                                                          0 );
                }
                if ( pszResolved == NULL )
				{
                    pszResolved = CPLSerializeXMLTree((CPLXMLNode *)pszNode);
					bToBeFreed = TRUE;
				}

                VSIFPrintfL ( fp, "        %s\n", pszResolved );
				if (bToBeFreed)
					CPLFree( pszResolved );
                VSIFPrintfL ( fp, "      </ResolvedGeometry>\n" );
                b_has_geom = TRUE;

                helper->iItemId += 1;
                i ++;
				pszNode = papszGeomList[i];
            }
        }
		gmlUpdateClass( pCC, poClass->GetElementName(), b_has_geom );
        iFeatureUID++;

        VSIFPrintfL ( fp, "    </%s>\n", poClass->GetElementName() ); 

        delete poFeature;
    }
    sqlite3_finalize ( hStmt );
    sqlite3_finalize ( hStmtNodes );

    VSIFPrintfL ( fp, "  </ResolvedTopoFeatureMembers>\n" ); 
    VSIFPrintfL ( fp, "</ResolvedTopoFeatureCollection>\n" ); 

    VSIFCloseL( fp );
		
    gmlUpdateFeatureClasses( pCC, pReader, m_bSequentialLayers );
    if ( *m_bSequentialLayers == TRUE )
        pReader->ReArrangeTemplateClasses( (void *)pCC );
    gmlDestroyClassContainer( pCC );
    return TRUE;
}

/**************************************************************/
/*                                                            */
/* private member(s):                                         */
/* any other funtion is implemented as "internal" static,     */
/* so to make all the SQLite own stuff nicely "invisible"     */
/*                                                            */
/**************************************************************/

int GMLReader::ReArrangeTemplateClasses ( const void * p_container )
{
/* rearranging the final FeatureClass list [SEQUENTIAL] */
    const struct huge_class_container * pCC = 
          (const struct huge_class_container *)p_container;
    int m_nSavedClassCount = GetClassCount();

/* saving the previous FeatureClass list */
    GMLFeatureClass **m_papoSavedClass = (GMLFeatureClass **) 
                    CPLMalloc( sizeof(void*) * m_nSavedClassCount );
    int clIdx;
    for (clIdx = 0; clIdx < GetClassCount(); clIdx++)
    {
    /* tranferring any previous FeatureClass */
        m_papoSavedClass[clIdx] = m_papoClass[clIdx];
    }

/* cleaning the previous FeatureClass list */
    SetClassListLocked( FALSE );
    CPLFree( m_papoClass );
    m_nClassCount = 0;
    m_papoClass = NULL;

    struct huge_class_item *pCI = pCC->pFirst;
    while ( pCI != NULL )
    { 
    /* 
    * re-inserting any required FeatureClassup
    * accordingly to actual SEQUENTIAL layout
    */
        GMLFeatureClass* poClass = NULL;
        for( int iClass = 0; iClass < m_nSavedClassCount; iClass++ )
        {
            GMLFeatureClass* poItem = m_papoSavedClass[iClass];
            if( EQUAL(poItem->GetName(), pCI->m_pszName) )
            {
                poClass = poItem;
                break;
            }
        }
        if (poClass != NULL)
        {
            if (poClass->GetFeatureCount() > 0)
                AddClass( poClass );
        }
        pCI = pCI->pNext;
    }
    SetClassListLocked( TRUE );
    
/* destroying the saved List and any unused FeatureClass */
    for( int iClass = 0; iClass < m_nSavedClassCount; iClass++ )
	{
		int bUnused = TRUE;
		GMLFeatureClass* poClass = m_papoSavedClass[iClass];
		for( int iClass2 = 0; iClass2 < m_nClassCount; iClass2++ )
		{
			if (m_papoClass[iClass2] == poClass)
            {
                bUnused = FALSE;
                break;
            }
        }
		if ( bUnused == TRUE )
			delete poClass;
	}
    CPLFree( m_papoSavedClass );
	return 1;
}
	
int GMLReader::ParseXMLHugeFile( const char *pszOutputFilename,
                                 const int bOutIsTempFile,
                                 const int bSqliteUnsafe )

{
    GMLFeature  *poFeature;
    int iFeatureUID = 0;
    int rc;
    sqlite3 *hDB = NULL;
    char *pszSQLiteFilename = NULL;
    struct huge_file_helper helper;
    char *pszErrMsg = NULL;

    /* initializing the helper struct */
    helper.hDB = NULL;
    helper.hNodes = NULL;
    helper.hStrings = NULL;
    helper.hHRefs = NULL;
    helper.iItemId = 0;
    helper.pFirst = NULL;
    helper.pLast = NULL;
	
/* -------------------------------------------------------------------- */
/*      Creating/Opening the SQLite DB file                             */
/* -------------------------------------------------------------------- */
    pszSQLiteFilename = CPLStrdup(
    CPLResetExtension( m_pszFilename, "sqlite" ) );
    VSIStatBufL statBufL;
    if ( VSIStatExL ( pszSQLiteFilename, &statBufL, VSI_STAT_EXISTS_FLAG) == 0)
    {
        CPLError( CE_Failure, CPLE_OpenFailed, 
                  "sqlite3_open(%s) failed:\n\tDB-file already exists", 
                  pszSQLiteFilename );
        CPLFree( pszSQLiteFilename );
        return FALSE;
    }

    rc = sqlite3_open( pszSQLiteFilename, &hDB );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_OpenFailed, 
                  "sqlite3_open(%s) failed: %s", 
                  pszSQLiteFilename, sqlite3_errmsg( hDB ) );
        return FALSE;
    }
    helper.hDB = hDB;
	
/* 
* setting SQLite for max speed; this is intrinsically unsafe,
* and the DB file could be potentially damaged (power failure ...]
*
* but after all this one simply is a TEMPORARY FILE !!
* so there is no real risk condition
*/
	if ( bSqliteUnsafe == TRUE )
	{
		rc = sqlite3_exec( hDB, "PRAGMA synchronous = OFF", NULL, NULL, &pszErrMsg );
		if( rc != SQLITE_OK )
		{
			CPLError( CE_Failure, CPLE_AppDefined,
					  "Unable to set PRAGMA synchronous = OFF: %s",
					  pszErrMsg );
			sqlite3_free( pszErrMsg );
		}
		rc = sqlite3_exec( hDB, "PRAGMA journal_mode = OFF", NULL, NULL, &pszErrMsg );
		if( rc != SQLITE_OK )
		{
			CPLError( CE_Failure, CPLE_AppDefined,
					  "Unable to set PRAGMA journal_mode = OFF: %s",
					  pszErrMsg );
			sqlite3_free( pszErrMsg );
		}
		rc = sqlite3_exec( hDB, "PRAGMA locking_mode = EXCLUSIVE", NULL, NULL, &pszErrMsg );
		if( rc != SQLITE_OK )
		{
			CPLError( CE_Failure, CPLE_AppDefined,
					  "Unable to set PRAGMA locking_mode = EXCLUSIVE: %s",
					  pszErrMsg );
			sqlite3_free( pszErrMsg );
		}
	}
	
    if( !SetupParser() )
    {
        gmlHugeFileCleanUp ( &helper );
        return FALSE;
    }

    /* creating SQLite tables and Insert cursors */
    if( gmlHugeFileSQLiteInit( &helper ) == FALSE )
    {
        gmlHugeFileCleanUp ( &helper );
        return FALSE;
    }

    /* processing GML features */
    while( (poFeature = NextFeature()) != NULL )
    {
        const CPLXMLNode* const * papszGeomList = poFeature->GetGeometryList();
        if (papszGeomList != NULL)
        {
            int i = 0;
            const CPLXMLNode *pszNode = papszGeomList[i];
            while(pszNode)
            {
                char * gmlText = CPLSerializeXMLTree((CPLXMLNode *)pszNode);
                if ( gmlHugeFileCheckXrefs( &helper, gmlText ) == FALSE )
                {
                    gmlHugeFileCleanUp ( &helper );
                    return FALSE;
                }
				CPLFree( gmlText );
                helper.iItemId += 1;
                i ++;
                pszNode = papszGeomList[i];
            }
        }
        iFeatureUID++;
        delete poFeature;
    }

    /* finalizing any SQLite Insert cursor */
    sqlite3_finalize ( helper.hNodes );
    helper.hNodes = NULL;
    sqlite3_finalize ( helper.hStrings );
    helper.hStrings = NULL;
    sqlite3_finalize ( helper.hHRefs );
    helper.hHRefs = NULL;
		
    /* confirming the still pending TRANSACTION */
    rc = sqlite3_exec( hDB, "COMMIT", NULL, NULL, &pszErrMsg );
    if( rc != SQLITE_OK )
    {
        CPLError( CE_Failure, CPLE_AppDefined, 
                  "Unable to perform COMMIT TRANSACTION: %s",
                  pszErrMsg );
        sqlite3_free( pszErrMsg );
        return FALSE;
    }

/* creating any further SQLite Index */
    if( gmlHugeFileSQLiteIndex( &helper ) == FALSE )
    {
        gmlHugeFileCleanUp ( &helper );
        return FALSE;
    }
	
/* attempting to resolve GML string */
    if( gmlHugeFileResolve( &helper ) == FALSE )
    {
        gmlHugeFileCleanUp ( &helper );
        return FALSE;
    }
	
    /* restarting the GML parser */
    if( !SetupParser() )
    {
        gmlHugeFileCleanUp ( &helper );
        return FALSE;
    }

    /* output: writing the revolved GML file */
    if ( gmlHugeFileWriteResolved( &helper, pszOutputFilename, this, 
	                               &m_bSequentialLayers ) == FALSE)
    {
        gmlHugeFileCleanUp ( &helper );
        return FALSE;
    }

    gmlHugeFileCleanUp ( &helper );
    if ( bOutIsTempFile == TRUE )
        VSIUnlink( pszSQLiteFilename );
    CPLFree( pszSQLiteFilename );
    return TRUE;
}

/**************************************************************/
/*                                                            */
/* an alternative <xlink:href> resolver based on SQLite       */
/*                                                            */
/**************************************************************/
int GMLReader::HugeFileResolver( const char *pszFile,
                                 int* pbOutIsTempFile,
                                 char **papszSkip,
                                 const int bSqliteUnsafe )

{
    // Check if the original source file is set.
    if( m_pszFilename == NULL )
    {
        CPLError( CE_Failure, CPLE_NotSupported,
        "GML source file needs to be set first with "
        "GMLReader::SetSourceFile()." );
        return FALSE;
    }
    if ( ParseXMLHugeFile( pszFile, *pbOutIsTempFile,
                           bSqliteUnsafe ) == FALSE )
        return FALSE;

    //set the source file to the resolved file
    CleanupParser();
    if (fpGML)
        VSIFCloseL(fpGML);
    fpGML = NULL;
    CPLFree( m_pszFilename );
    m_pszFilename = CPLStrdup( pszFile );
    return TRUE;
}
